import logging
import os
import shutil
import sys
import tarfile
import urllib.request
import warnings
from collections.abc import Iterable, Sequence
from typing import Callable

import pytest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
import suricata_check
from click.testing import CliRunner

_regex_provider = suricata_check.utils.regex.get_regex_provider()

ET_OPEN_URLS = {
    "v5": "https://rules.emergingthreats.net/open/suricata-5.0/emerging-all.rules.tar.gz",
    "v7": "https://rules.emergingthreats.net/open/suricata-7.0.3/emerging-all.rules.tar.gz",
}
SNORT_COMMUNITY_URL = (
    "https://www.snort.org/downloads/community/snort3-community-rules.tar.gz"
)


@pytest.fixture(autouse=True)
def __run_around_tests():
    # Clean up from previous tests.
    if os.path.exists("tests/data/out") and os.path.isdir("tests/data/out"):
        for f in os.listdir("tests/data/out"):
            os.remove(os.path.join("tests/data/out", f))

    yield

    # Optionally clean up after the test run.
    logging.shutdown()


@pytest.mark.serial()
def test_main():
    os.environ["SURICATA_CHECK_FORCE_LOGGING"] = "TRUE"
    runner = CliRunner()
    result = runner.invoke(
        suricata_check.main,
        ("--rules=tests/data/test.rules", "--out=tests/data/out", "--log-level=DEBUG"),
        catch_exceptions=False,
    )

    __check_log_file()

    if result.exit_code != 0:
        pytest.fail(result.output)


@pytest.mark.serial()
def test_main_single_rule():
    os.environ["SURICATA_CHECK_FORCE_LOGGING"] = "TRUE"
    runner = CliRunner()
    result = runner.invoke(
        suricata_check.main,
        (
            """--single-rule=alert ip $HOME_NET any -> $EXTERNAL_NET any (msg:"Test"; sid:1;)""",
            "--out=tests/data/out",
            "--log-level=DEBUG",
        ),
        catch_exceptions=False,
    )

    __check_log_file()

    if result.exit_code != 0:
        pytest.fail(result.output)


@pytest.mark.serial()
def test_main_gitlab():
    os.environ["SURICATA_CHECK_FORCE_LOGGING"] = "TRUE"
    runner = CliRunner()
    result = runner.invoke(
        suricata_check.main,
        (
            "--rules=tests/data/test.rules",
            "--out=tests/data/out",
            "--log-level=DEBUG",
            "--gitlab",
        ),
        catch_exceptions=False,
    )

    __check_log_file()

    if result.exit_code != 0:
        pytest.fail(result.output)


@pytest.mark.serial()
def test_main_github():
    os.environ["SURICATA_CHECK_FORCE_LOGGING"] = "TRUE"
    runner = CliRunner()
    result = runner.invoke(
        suricata_check.main,
        (
            "--rules=tests/data/test.rules",
            "--out=tests/data/out",
            "--log-level=DEBUG",
            "--github",
        ),
        catch_exceptions=False,
    )

    __check_log_file()

    if result.exit_code != 0:
        pytest.fail(result.output)


@pytest.mark.serial()
def test_main_evaluate_disabled():
    os.environ["SURICATA_CHECK_FORCE_LOGGING"] = "TRUE"
    runner = CliRunner()
    result = runner.invoke(
        suricata_check.main,
        (
            "--rules=tests/data/test.rules",
            "--out=tests/data/out",
            "--log-level=DEBUG",
            "--evaluate-disabled",
        ),
        catch_exceptions=False,
    )

    __check_log_file()

    if result.exit_code != 0:
        pytest.fail(result.output)


@pytest.mark.serial()
def test_main_issue_severity():
    os.environ["SURICATA_CHECK_FORCE_LOGGING"] = "TRUE"
    runner = CliRunner()
    result = runner.invoke(
        suricata_check.main,
        (
            "--rules=tests/data/test.rules",
            "--out=tests/data/out",
            "--log-level=DEBUG",
            "--issue-severity=WARNING",
        ),
        catch_exceptions=False,
    )

    __check_log_file()

    if result.exit_code != 0:
        pytest.fail(result.output)


@pytest.mark.serial()
def test_main_include_all():
    os.environ["SURICATA_CHECK_FORCE_LOGGING"] = "TRUE"
    runner = CliRunner()
    result = runner.invoke(
        suricata_check.main,
        (
            "--rules=tests/data/test.rules",
            "--out=tests/data/out",
            "--log-level=DEBUG",
            "--include-all",
        ),
        catch_exceptions=False,
    )

    __check_log_file()

    if result.exit_code != 0:
        pytest.fail(result.output)


@pytest.mark.serial()
def test_main_include():
    os.environ["SURICATA_CHECK_FORCE_LOGGING"] = "TRUE"
    runner = CliRunner()
    result = runner.invoke(
        suricata_check.main,
        (
            "--rules=tests/data/test.rules",
            "--out=tests/data/out",
            "--log-level=DEBUG",
            "--include=M.*",
        ),
        catch_exceptions=False,
    )

    __check_log_file()

    if result.exit_code != 0:
        pytest.fail(result.output)


@pytest.mark.serial()
def test_main_exclude():
    os.environ["SURICATA_CHECK_FORCE_LOGGING"] = "TRUE"
    runner = CliRunner()
    result = runner.invoke(
        suricata_check.main,
        (
            "--rules=tests/data/test.rules",
            "--out=tests/data/out",
            "--log-level=DEBUG",
            "--exclude=M.*",
        ),
        catch_exceptions=False,
    )

    __check_log_file()

    if result.exit_code != 0:
        pytest.fail(result.output)


@pytest.mark.slow()
@pytest.mark.serial()
@pytest.hookimpl(trylast=True)
@pytest.mark.parametrize(("version", "et_open_url"), ET_OPEN_URLS.items())
def test_main_integration_et_open(version, et_open_url):
    os.environ["SURICATA_CHECK_FORCE_LOGGING"] = "TRUE"

    # Retrieve the latest ET Open rules if not present.
    if not os.path.exists(f"tests/data/emerging-all-{version}.rules"):
        if not os.path.exists(f"tests/data/emerging-all-{version}.rules.tar.gz"):
            urllib.request.urlretrieve(
                et_open_url,
                f"tests/data/emerging-all-{version}.rules.tar.gz",
            )

        tarfile.open(f"tests/data/emerging-all-{version}.rules.tar.gz").extract(
            "emerging-all.rules",
            "tests/data/temp",
        )
        os.remove(f"tests/data/emerging-all-{version}.rules.tar.gz")
        shutil.move(
            "tests/data/temp/emerging-all.rules",
            f"tests/data/emerging-all-{version}.rules",
        )
        shutil.rmtree("tests/data/temp")

    runner = CliRunner()
    result = runner.invoke(
        suricata_check.main,
        (
            f"--rules=tests/data/emerging-all-{version}.rules",
            "--out=tests/data/out",
            "--log-level=INFO",
        ),
        catch_exceptions=False,
    )

    __check_log_file()

    if result.exit_code != 0:
        pytest.fail(result.output)


@pytest.mark.slow()
@pytest.mark.serial()
@pytest.hookimpl(trylast=True)
def test_main_integration_snort_community():
    # Retrieve the latest Snort rules if not present.
    if not os.path.exists("tests/data/snort3-community.rules"):
        if not os.path.exists("tests/data/snort3-community-rules.tar.gz"):
            urllib.request.urlretrieve(
                SNORT_COMMUNITY_URL,
                "tests/data/snort3-community-rules.tar.gz",
            )

        tarfile.open("tests/data/snort3-community-rules.tar.gz").extract(
            "snort3-community-rules/snort3-community.rules",
            "tests/data/temp",
        )
        os.remove("tests/data/snort3-community-rules.tar.gz")
        shutil.move(
            "tests/data/temp/snort3-community-rules/snort3-community.rules",
            "tests/data/snort3-community.rules",
        )
        shutil.rmtree("tests/data/temp")

    runner = CliRunner()
    result = runner.invoke(
        suricata_check.main,
        (
            "--rules=tests/data/snort3-community.rules",
            "--out=tests/data/out",
            "--log-level=INFO",
        ),
        catch_exceptions=False,
    )

    if result.exit_code != 0:
        pytest.fail(result.output)

    # We do not check the log file as we know some Snort rules are invalid Suricata rules.


@pytest.mark.serial()
def test_main_ini():
    os.environ["SURICATA_CHECK_FORCE_LOGGING"] = "TRUE"
    runner = CliRunner()
    result = runner.invoke(
        suricata_check.main,
        (
            "--rules=tests/data/test.rules",
            "--out=tests/data/out",
            "--log-level=DEBUG",
            "--ini=tests/data/suricata-check.ini",
        ),
        catch_exceptions=False,
    )

    __check_log_file()

    if result.exit_code != 0:
        pytest.fail(result.output)


@pytest.mark.serial()
def test_main_error():
    os.environ["SURICATA_CHECK_FORCE_LOGGING"] = "TRUE"
    runner = CliRunner()
    result = runner.invoke(
        suricata_check.main,
        (
            "--rules=tests/data/test_error.rules",
            "--out=tests/data/out",
            "--log-level=DEBUG",
        ),
        catch_exceptions=False,
    )

    if result.exit_code != 0:
        pytest.fail(result.output)


@pytest.mark.serial()
def test_main_ignore():
    os.environ["SURICATA_CHECK_FORCE_LOGGING"] = "TRUE"
    runner = CliRunner()
    result = runner.invoke(
        suricata_check.main,
        (
            "--rules=tests/data/test_ignore.rules",
            "--out=tests/data/out",
            "--log-level=DEBUG",
        ),
        catch_exceptions=False,
    )

    __check_log_file()
    __check_fast_file([lambda line: "M001" not in line], [lambda file: "M000" in file])

    if result.exit_code != 0:
        pytest.fail(result.output)


def test_get_ini_exclude_tuple():
    ini = os.path.abspath(os.path.join("tests", "data", "suricata-check.ini"))
    ini_kwargs = suricata_check.get_ini_kwargs(ini)

    assert "exclude" in ini_kwargs
    assert isinstance(ini_kwargs["exclude"], tuple)
    for pattern in ("P.*", "Q.*"):
        assert isinstance(pattern, str)
    assert ini_kwargs["exclude"] == ("P.*", "Q.*")


def test_get_ini_issue_severity_str():
    ini = os.path.abspath(os.path.join("tests", "data", "suricata-check.ini"))
    ini_kwargs = suricata_check.get_ini_kwargs(ini)

    assert "issue_severity" in ini_kwargs
    assert isinstance(ini_kwargs["issue_severity"], str)
    assert not ini_kwargs["issue_severity"].startswith('"')
    assert not ini_kwargs["issue_severity"].endswith('"')
    assert ini_kwargs["issue_severity"] == "INFO"


def test_get_checkers():
    logging.basicConfig(level=logging.DEBUG)
    suricata_check.get_checkers()


def test_get_checkers_multiple_include():
    logging.basicConfig(level=logging.DEBUG)
    assert len(suricata_check.get_checkers(include=("M.*", "S.*"))) > 0


def test_get_checkers_include():
    logging.basicConfig(level=logging.DEBUG)
    assert len(suricata_check.get_checkers(include=("M.*",))) == 1


def test_get_checkers_exclude():
    logging.basicConfig(level=logging.DEBUG)
    assert len(suricata_check.get_checkers(exclude=("(?!M).*",))) == 1


def test_analyze_rule():
    logging.basicConfig(level=logging.DEBUG)
    rule = suricata_check.rule.parse(
        """alert ip $HOME_NET any -> $EXTERNAL_NET any (msg:"Test"; sid:1;)""",
    )
    assert rule is not None

    suricata_check.analyze_rule(rule)


def test_version():
    logging.basicConfig(level=logging.DEBUG)
    if not hasattr(suricata_check, "__version__"):
        pytest.fail("suricata_check has no attribute __version__")
    from suricata_check._version import __version__  # noqa: RUF100, PLC0415

    if __version__ == "unknown":
        warnings.warn(RuntimeWarning("Version is unknown."))


def __check_log_file(checks: Iterable[Callable[[str], bool]] = []):
    log_file = "tests/data/out/suricata-check.log"

    if not os.path.exists(log_file):
        pytest.fail("No log file found.")
        return

    with open(log_file) as log_fh:
        for line in log_fh.readlines():
            if _regex_provider.match(
                r".+ - .+ - (ERROR|CRITICAL) - .+(?<!Error parsing rule)",
                line,
            ):
                pytest.fail(line)
            for check in checks:
                if not check(line):
                    pytest.fail(line)
            if _regex_provider.match(r".+ - .+ - (WARNING) - .+", line):
                warnings.warn(RuntimeWarning(line))


def __check_fast_file(
    line_checks: Sequence[Callable[[str], bool]] = [],
    file_checks: Sequence[Callable[[str], bool]] = [],
):
    log_file = "tests/data/out/suricata-check-fast.log"

    if not os.path.exists(log_file):
        pytest.fail("No fast file found.")
        return

    if len(line_checks) > 0:
        with open(log_file) as log_fh:
            for line in log_fh.readlines():
                for check in line_checks:
                    if not check(line):
                        pytest.fail(line)

    if len(file_checks) > 0:
        with open(log_file) as log_fh:
            file = "\n".join(log_fh.readlines())
            for check in file_checks:
                if not check(file):
                    pytest.fail(file)


def __main__():
    pytest.main()



import gettext
import logging
from typing import Any, Callable, TypeVar

import click

from suricata_check._version import get_version

_AnyCallable = Callable[..., Any]
FC = TypeVar("FC", bound="_AnyCallable | click.Command")


def help_option(*param_decls: str, **kwargs: dict[str, Any]) -> Callable[[FC], FC]:
    """``--help`` option which immediately printsversion help info and exits.

    :param param_decls: One or more option names. Defaults to the single
        value ``"--help"``.
    :param kwargs: Extra arguments are passed to :func:`option`.
    """

    def show_help(
        ctx: click.Context, param: click.Parameter, value: bool  # noqa: ARG001
    ) -> None:
        """Callback that print the help page on ``<stdout>`` and exits."""
        click.echo("suricata-check {}\n".format(get_version()))

        if value and not ctx.resilient_parsing:
            click.echo(ctx.get_help(), color=ctx.color)
            ctx.exit()

    if not param_decls:
        param_decls = ("--help",)

    kwargs.setdefault("is_flag", True)  # pyright: ignore[reportArgumentType]
    kwargs.setdefault("expose_value", False)  # pyright: ignore[reportArgumentType]
    kwargs.setdefault("is_eager", True)  # pyright: ignore[reportArgumentType]
    kwargs.setdefault(
        "help",
        gettext.gettext(
            "Show this message and exit."
        ),  # pyright: ignore[reportArgumentType]
    )
    kwargs.setdefault("callback", show_help)  # pyright: ignore[reportArgumentType]

    return click.option(*param_decls, **kwargs)  # pyright: ignore[reportArgumentType]


class ClickHandler(logging.Handler):
    """Handler to color and write logging messages for the click module."""

    def __init__(
        self: "ClickHandler",
        level: int = 0,
        github: bool = False,
        github_level: int = logging.WARNING,
        **kwargs: dict,
    ) -> None:
        super().__init__(level, **kwargs)
        self.github = github
        self.github_level = github_level

    def emit(self: "ClickHandler", record: logging.LogRecord) -> None:
        """Log the record via click stdout with appropriate colors."""
        msg = self.format(record)

        if logging.getLevelName(record.levelno) == "DEBUG":
            click.secho(msg, color=True, dim=True)
        if logging.getLevelName(record.levelno) == "INFO":
            click.secho(msg, color=True)
        if logging.getLevelName(record.levelno) == "WARNING":
            click.secho(msg, color=True, bold=True, fg="yellow")
        if logging.getLevelName(record.levelno) == "ERROR":
            click.secho(msg, color=True, bold=True, fg="red")
        if logging.getLevelName(record.levelno) == "CRITICAL":
            click.secho(msg, color=True, bold=True, blink=True, fg="red")

        if self.github and record.levelno >= self.github_level:
            print(f"::debug::{msg}")  # noqa: T201



"""`MandatoryChecker`."""

import logging

from suricata_check.checkers.interface import CheckerInterface
from suricata_check.utils.checker import is_rule_option_set
from suricata_check.utils.checker_typing import ISSUES_TYPE, Issue, Rule


class MandatoryChecker(CheckerInterface):
    """The `MandatoryChecker` contains several checks based on the Suricata syntax that are critical.

    Codes M000-M009 report on missing mandatory rule options.
    """

    codes = {
        "M000": {"severity": logging.ERROR},
        "M001": {"severity": logging.ERROR},
    }

    def _check_rule(
        self: "MandatoryChecker",
        rule: Rule,
    ) -> ISSUES_TYPE:
        issues: ISSUES_TYPE = []

        if not is_rule_option_set(rule, "msg"):
            issues.append(
                Issue(
                    code="M000",
                    message="The rule did not specify a msg, which is a mandatory field.",
                )
            )

        if not is_rule_option_set(rule, "sid"):
            issues.append(
                Issue(
                    code="M001",
                    message="The rule did not specify a sid, which is a mandatory field.",
                )
            )

        return issues



import datetime
import json
import logging
import os
import re
import subprocess
from functools import lru_cache
from importlib.metadata import PackageNotFoundError, requires, version
from typing import Optional

import requests
from packaging.version import Version

SURICATA_CHECK_DIR = os.path.dirname(__file__)
UPDATE_CHECK_CACHE_PATH = os.path.expanduser("~/.suricata_check_version_check.json")

_logger = logging.getLogger(__name__)


def __get_git_revision_short_hash() -> str:
    return (
        subprocess.check_output(["git", "rev-parse", "--short", "HEAD"])
        .decode("ascii")
        .strip()
    )


def get_version() -> str:
    v = "unknown"

    git_dir = os.path.join(SURICATA_CHECK_DIR, "..", ".git")
    if os.path.exists(git_dir):
        try:
            import setuptools_git_versioning  # noqa: RUF100, PLC0415

            v = str(
                setuptools_git_versioning.get_version(
                    root=os.path.join(SURICATA_CHECK_DIR, "..")
                )
            )
            _logger.debug(
                "Detected suricata-check version using setuptools_git_versioning: %s", v
            )
        except:  # noqa: E722
            v = __get_git_revision_short_hash()
            _logger.debug("Detected suricata-check version using git: %s", v)
    else:
        try:
            v = version("suricata-check")
            _logger.debug("Detected suricata-check version using importlib: %s", v)
        except PackageNotFoundError:
            _logger.debug("Failed to detect suricata-check version: %s", v)

    return v


__version__: str = get_version()

__user_agent = f"suricata-check/{__version__} (+https://suricata-check.teuwen.net/)"


def get_dependency_versions() -> dict:
    d = {}

    requirements = None
    try:
        requirements = requires("suricata-check")
        _logger.debug("Detected suricata-check requirements using importlib")
    except PackageNotFoundError:
        requirements_path = os.path.join(SURICATA_CHECK_DIR, "..", "requirements.txt")
        if os.path.exists(requirements_path):
            with open(requirements_path) as fh:
                requirements = fh.readlines()
                requirements = filter(
                    lambda x: len(x.strip()) == 0 or x.strip().startswith("#"),
                    requirements,
                )

            _logger.debug("Detected suricata-check requirements using requirements.txt")

    if requirements is None:
        _logger.debug("Failed to detect suricata-check requirements")
        return d

    for requirement in requirements:
        match = re.compile(r"""^([^=<>]+)(.*)$""").match(requirement)
        if match is None:
            _logger.debug("Failed to parse requirement: %s", requirement)
            continue
        required_package, _ = match.groups()
        try:
            d[required_package] = version(required_package)
        except PackageNotFoundError:
            d[required_package] = "unknown"

    return d


def __get_latest_version() -> Optional[str]:
    cached_data = __get_saved_check_update()
    headers = {"User-Agent": __user_agent}
    if cached_data is not None:
        if (
            "response_headers" in cached_data
            and "etag" in cached_data["response_headers"]
        ):
            headers["If-None-Match"] = cached_data["response_headers"]["etag"]
        if "last_checked" in cached_data:
            headers["If-Modified-Since"] = datetime.datetime.fromisoformat(
                cached_data["last_checked"]
            ).strftime("%a, %d %b %Y %H:%M:%S GMT")

    try:
        response = requests.get(
            "https://pypi.org/pypi/suricata-check/json", headers=headers, timeout=5
        )

        if response.status_code == requests.codes.ok:
            pypi_json = response.json()
            __save_check_update(
                pypi_json, {k.lower(): v for k, v in response.headers.items()}
            )
            return pypi_json["info"]["version"]

        if response.status_code == requests.codes.not_modified:
            assert cached_data is not None
            _logger.debug("Using cached PyPI response data for update check.")
            return cached_data["pypi_json"]["info"]["version"]
    except requests.RequestException:
        _logger.warning("Failed to perform update check.")
    return None


@lru_cache(maxsize=1)
def __get_saved_check_update() -> Optional[dict]:
    if not os.path.exists(UPDATE_CHECK_CACHE_PATH):
        return None

    try:
        with open(UPDATE_CHECK_CACHE_PATH, "r") as f:
            data = json.load(f)
    except OSError:
        _logger.warning("Failed to read last date version was checked from cache file.")
        os.remove(UPDATE_CHECK_CACHE_PATH)
        return None
    except json.JSONDecodeError:
        _logger.warning(
            "Failed to decode cache file to determine last date version was checked."
        )
        return None

    if not isinstance(data, dict):
        _logger.warning(
            "Cache file documenting the last date version was checked is malformed."
        )
        os.remove(UPDATE_CHECK_CACHE_PATH)
        return None

    return data


def __should_check_update() -> bool:
    current_version = __version__
    if current_version == "unknown":
        _logger.warning(
            "Skipping update check because current version cannot be determined."
        )
        return False
    if "dirty" in current_version:
        _logger.warning("Skipping update check because local changes are detected.")
        return False

    if not os.path.exists(UPDATE_CHECK_CACHE_PATH):
        return True

    data = __get_saved_check_update()
    if data is None:
        return True

    try:
        last_checked = datetime.datetime.fromisoformat(data["last_checked"])
        if (datetime.datetime.now(tz=datetime.timezone.utc) - last_checked).days < 1:
            return False
    except KeyError:
        _logger.warning(
            "Cache file documenting the last date version was checked is malformed."
        )

    return True


def __save_check_update(pypi_json: dict, response_headers: dict) -> None:
    try:
        with open(UPDATE_CHECK_CACHE_PATH, "w") as f:
            json.dump(
                {
                    "last_checked": datetime.datetime.now(
                        tz=datetime.timezone.utc
                    ).isoformat(),
                    "pypi_json": pypi_json,
                    "response_headers": response_headers,
                },
                f,
            )
    except OSError:
        _logger.warning("Failed to write current date to cache file for update checks.")


def check_for_update() -> None:
    if not __should_check_update():
        return

    current_version = __version__
    latest_version = __get_latest_version()

    if latest_version is None:
        _logger.warning("Failed to check for updates of suricata-check.")
        return

    if Version(latest_version) > Version(current_version):
        _logger.warning(
            "A new version of suricata-check is available: %s (you have %s)",
            latest_version,
            current_version,
        )
        _logger.warning("Run `pip install --upgrade suricata-check` to update.")
        _logger.warning(
            "You can find the full changelog of what has changed here: %s",
            "https://github.com/Koen1999/suricata-check/releases",
        )
        return

    _logger.info(
        "You are using the latest version of suricata-check (%s).", __version__
    )



"""`WhitespaceChecker`."""

import logging

from suricata_check.checkers.interface import CheckerInterface
from suricata_check.utils.checker import (
    is_rule_option_equal_to_regex,
)
from suricata_check.utils.checker_typing import ISSUES_TYPE, Issue, Rule
from suricata_check.utils.regex import (
    HEADER_REGEX,
    get_regex_provider,
)

_regex_provider = get_regex_provider()

# Regular expressions are placed here such that they are compiled only once.
# This has a significant impact on the performance.
REGEX_S100 = _regex_provider.compile(
    rf"^(\s*#)?\s*{HEADER_REGEX.pattern}\s*\( .*\)\s*(#.*)?$",
)
REGEX_S101 = _regex_provider.compile(
    rf"^(\s*#)?\s*{HEADER_REGEX.pattern}\s*\(.* \)\s*(#.*)?$",
)
REGEX_S102 = _regex_provider.compile(
    rf"^(\s*#)?\s*{HEADER_REGEX.pattern}\s*\(.+ :.+\)\s*(#.*)?$",
)
REGEX_S103 = _regex_provider.compile(
    rf"^(\s*#)?\s*{HEADER_REGEX.pattern}\s*\(.+: .+\)\s*(#.*)?$",
)
REGEX_S104 = _regex_provider.compile(
    rf"^(\s*#)?\s*{HEADER_REGEX.pattern}\s*\(.+ ;.+\)\s*(#.*)?$",
)
REGEX_S105 = _regex_provider.compile(
    rf"^(\s*#)?\s*{HEADER_REGEX.pattern}\s*\(.+; \s+.+\)\s*(#.*)?$",
)
REGEX_S106 = _regex_provider.compile(r'^".*\|.*  .*\|.*"$')
REGEX_S110 = _regex_provider.compile(
    rf"^(\s*#)?\s*{HEADER_REGEX.pattern}\s*\(.+;(?! ).+\)\s*(#.*)?$",
)
REGEX_S111 = _regex_provider.compile(r'^".*\|.*[a-fA-F0-9]{4}.*\|.*"$')
REGEX_S120 = _regex_provider.compile(
    r'^"([^\|]*|(\|[\sa-zA-Z0-9]*\|))*(\\?[\x3a\x3b\x20\x27\x7b\x5c\x2f\x60\x24\x28\x29]+|\\[\x22\x7c]+)([^\|]*|(\|[\sa-zA-Z0-9]*\|))*"$',
)
REGEX_S121 = _regex_provider.compile(
    r"^\"/.*(\\?[\x3a\x3b\x20\x22\x27\x2f\x60]+|\\[\x7b\x5c\x7c\x24\x28\x29]+).*/[ism]*\"$",
)
REGEX_S122 = _regex_provider.compile(r'^".*\\.*"$')
REGEX_S123 = _regex_provider.compile(
    r'^".*(?!\\(a|c[0-127]|e|f|n|r|t|0[0-9]{2}|[0-9]{3}|0\{[0-9]{3}\}|x[0-9a-f]{2}|x[0-9a-f]{3}|u[0-9a-f]{4}|d|D|h|H|s|S|v|V|w|W))(\\.).*"$'
)


class WhitespaceChecker(CheckerInterface):
    """The `WhitespaceChecker` contains several checks based on the Suricata Style guide relating to formatting rules.

    Codes S100-S109 report on unneccessary whitespace that should be removed.

    Codes S110-S119 report on missing whitespace that should be added.

    Codes S120-S129 report on non-standard escaping of special characters.
    """

    codes = {
        "S100": {"severity": logging.INFO},
        "S101": {"severity": logging.INFO},
        "S102": {"severity": logging.INFO},
        "S103": {"severity": logging.INFO},
        "S104": {"severity": logging.INFO},
        "S105": {"severity": logging.INFO},
        "S106": {"severity": logging.INFO},
        "S110": {"severity": logging.INFO},
        "S111": {"severity": logging.INFO},
        "S120": {"severity": logging.INFO},
        "S121": {"severity": logging.INFO},
        "S122": {"severity": logging.INFO},
        "S123": {"severity": logging.INFO},
    }

    def _check_rule(  # noqa: C901, PLR0912
        self: "WhitespaceChecker",
        rule: Rule,
    ) -> ISSUES_TYPE:
        issues: ISSUES_TYPE = []

        if (
            REGEX_S100.match(
                rule["raw"].strip(),
            )
            is not None
        ):
            issues.append(
                Issue(
                    code="S100",
                    message="""The rule contains unneccessary whitespace after opening the rule body with.
Consider removing the unneccessary whitespace.""",
                ),
            )

        if (
            REGEX_S101.match(
                rule["raw"].strip(),
            )
            is not None
        ):
            issues.append(
                Issue(
                    code="S101",
                    message="""The rule contains unneccessary whitespace before closing the rule body with.
Consider removing the unneccessary whitespace.""",
                ),
            )

        if (
            REGEX_S102.match(
                rule["raw"].strip(),
            )
            is not None
        ):
            issues.append(
                Issue(
                    code="S102",
                    message="""The rule contains unneccessary whitespace before the colon (:) after an option name.
Consider removing the unneccessary whitespace.""",
                ),
            )

        if (
            REGEX_S103.match(
                rule["raw"].strip(),
            )
            is not None
        ):
            issues.append(
                Issue(
                    code="S103",
                    message="""The rule contains unneccessary whitespace before the colon (:) after an option name.
Consider removing the unneccessary whitespace.""",
                ),
            )

        if (
            REGEX_S104.match(
                rule["raw"].strip(),
            )
            is not None
        ):
            issues.append(
                Issue(
                    code="S104",
                    message="""The rule contains unneccessary whitespace before the semicolon (;) after an option value.
Consider removing the unneccessary whitespace.""",
                ),
            )

        if (
            REGEX_S105.match(
                rule["raw"].strip(),
            )
            is not None
        ):
            issues.append(
                Issue(
                    code="S105",
                    message="""The rule contains more than one space between options after an option value.
Consider replacing the unneccessary whitespace by a single space.""",
                ),
            )

        if is_rule_option_equal_to_regex(
            rule,
            "content",
            REGEX_S106,
        ):
            issues.append(
                Issue(
                    code="S106",
                    message="""The rule contains more than one space between bytes in content.
Consider replacing the unneccessary whitespace by a single space.""",
                ),
            )

        if (
            REGEX_S110.match(
                rule["raw"].strip(),
            )
            is not None
        ):
            issues.append(
                Issue(
                    code="S110",
                    message="""The rule does not contain a space between the end of after an option value.
Consider adding a single space.""",
                ),
            )

        if is_rule_option_equal_to_regex(
            rule,
            "content",
            REGEX_S111,
        ):
            issues.append(
                Issue(
                    code="S111",
                    message="""The rule contains more than no spaces between bytes in content.
Consider replacing adding a single space.""",
                ),
            )

        if is_rule_option_equal_to_regex(
            rule,
            "content",
            REGEX_S120,
        ):
            issues.append(
                Issue(
                    code="S120",
                    message="""The rule did not escape \
(\\x3a\\x3b\\x20\\x22\\x27\\x7b\\x7c\\x5c\\x2f\\x60\\x24\\x28\\x29) in a content field.
Consider using hex encoding instead.""",
                ),
            )

        if is_rule_option_equal_to_regex(
            rule,
            "pcre",
            REGEX_S121,
        ):
            issues.append(
                Issue(
                    code="S121",
                    message="""The rule did escape \
(\\x3a\\x3b\\x20\\x22\\x27\\x7b\\x7c\\x5c\\x2f\\x60\\x24\\x28\\x29) in a pcre field.
Consider using hex encoding instead.""",
                ),
            )

        if is_rule_option_equal_to_regex(rule, "content", REGEX_S122):
            issues.append(
                Issue(
                    code="S122",
                    message="""The rule escaped special characters using a blackslash (\\) in a content field.
Consider using hex encoding instead.""",
                ),
            )

        if is_rule_option_equal_to_regex(rule, "pcre", REGEX_S123):
            issues.append(
                Issue(
                    code="S123",
                    message="""The rule escaped special characters using a blackslash (\\) in a pcre field.
Consider using hex encoding instead.""",
                ),
            )

        return issues



"""`BestChecker`."""

import logging

from suricata_check.checkers.interface import CheckerInterface
from suricata_check.utils.checker import (
    get_rule_option,
    is_rule_option_set,
    is_rule_suboption_set,
)
from suricata_check.utils.checker_typing import ISSUES_TYPE, Issue, Rule


class BestChecker(CheckerInterface):
    """The `BestChecker` contains several checks for best practices to improve the experience of Suricata rules for everyone.

    Codes C100-C110 report on missing fields that should be set.
    """

    codes = {
        "C100": {"severity": logging.INFO},
        "C101": {"severity": logging.INFO},
        "C102": {"severity": logging.INFO},
    }

    def _check_rule(
        self: "BestChecker",
        rule: Rule,
    ) -> ISSUES_TYPE:
        issues: ISSUES_TYPE = []

        if not (
            is_rule_option_set(rule, "noalert")
            or is_rule_suboption_set(rule, "flowbits", "noalert")
        ) and not is_rule_option_set(rule, "target"):
            issues.append(
                Issue(
                    code="C100",
                    message="""\
The rule does not use the `target` Suricata meta option.
Consider adding the `target` option to specify which IP address is the target of the attack.\
""",
                ),
            )

        if not is_rule_suboption_set(rule, "metadata", "created_at"):
            issues.append(
                Issue(
                    code="C101",
                    message="""\
The rule does not use set the `created_at` metadata option.
Consider adding the `created_at` metadata option to inform users of the recency of this signature.\
""",
                ),
            )

        if (
            is_rule_option_set(rule, "rev")
            and int(get_rule_option(rule, "rev")) > 1  # type: ignore reportArgumentType
            and not is_rule_suboption_set(rule, "metadata", "updated_at")
        ):
            issues.append(
                Issue(
                    code="C102",
                    message="""\
The rule does not use set the `updated_at` metadata option while it has been revised since creation.
Consider adding the `updated_at` metadata option to inform users of the recency of this signature.\
""",
                ),
            )

        return issues



"""`suricata_check` is a module and command line utility to provide feedback on Suricata rules."""

from suricata_check import checkers, rule, tests, utils
from suricata_check._checkers import get_checkers
from suricata_check._version import (
    SURICATA_CHECK_DIR,
    __version__,
    check_for_update,
    get_dependency_versions,
)
from suricata_check.suricata_check import (
    analyze_rule,
    get_ini_kwargs,
    main,
    process_rules_file,
)

__all__ = (
    "SURICATA_CHECK_DIR",
    "__version__",
    "analyze_rule",
    "check_for_update",
    "checkers",
    "get_checkers",
    "get_dependency_versions",
    "get_ini_kwargs",
    "main",
    "process_rules_file",
    "rule",
    "tests",
    "utils",
)
